/* Copyright 2009, UCAR/Unidata and OPeNDAP, Inc.
   See the COPYRIGHT file for more information. */

#ifdef HAVE_UNISTD_H
#include <unistd.h>
#endif
#include "ocinternal.h"
#include "ocdata.h"
#include "occontent.h"
#include "ocdebug.h"

/* define a large stack of DUMPPATH instances */
static struct DUMPPATH {
    OCnode* node;
    OCmode  mode;
    unsigned int index;
    char*  fieldname;
} stack[1024];
static unsigned int stacknext;

static void
pushstack(OCnode* node, OCmode mode, unsigned int index, char* fieldname)
{
   stack[stacknext].node = node;
   stack[stacknext].mode = mode;
   stack[stacknext].index = index;
   stack[stacknext].fieldname = fieldname;
   stacknext++;
}

static void popstack() {stacknext--;}

/* Command line options */
int verbose; /* -v */
int debug; /* -d | -Dn */
int inmemory; /* -m */
int optdas; /* -p */
int optdds; /* -p */
int optdatadds; /* -p */
int generate; /* -g */
char* projection; /* -P */
char* selection; /* -S */
OClist* userparams; /* -U */
int showattributes; /* -M */
int linear; /* -L */
char* fileurl;

extern int optind;
extern int opterr;
extern char* optarg;

/* Forward*/
static void usage(char*);
static int fail(char*);
static void check_err(int stat);
static void dumpflags(void);

static void processdata(char*);
static void printleaves(OCconnection*, OCnode* node, OCcontent*);
static void printleavesr(OCconnection*, OCnode* node, OCcontent*,OCbytes*);
static void dumpinstance(OCnode* node, size_t count, void* memory, OCbytes*);
static char* generatedds(OCconnection* state);
static void generateddsr(OCconnection*, OCnode*, OCbytes*, int depth);
static char* generatedas(OCconnection* state);
static void generatedasr(OCconnection*, OCnode*, OCbytes*, int depth);
static void generateattributes(OCnode* node, OCbytes*, int);
static char* stringescape(char*);
static char* idescape(char*);

int
main(int argc, char **argv)
{
    int c;

    verbose=0; /* -v */
    debug=0; /* -d | -Dn */
    inmemory=0; /* -m */
    optdas=0; /* -p */
    optdds=0; /* -p */
    optdatadds=0; /* -p */
    generate=0; /* -g */
    linear=0; /* -L */
    showattributes=0; /* -M */
    projection=NULL; /* -P */
    selection=NULL; /* -S */
    userparams=oclistnew(); /* -U */
    fileurl=NULL;

    opterr = 1;
    while ((c = getopt(argc, argv, "vdgmtMLp:D:P:S:U:")) != EOF) {
        switch (c) {
        case 'v': verbose = 1; break;
        case 'd': debug = 1; break;
        case 'D': debug = atoi(optarg); break;
        case 'g': generate = 1; break;
	case 'L': linear = 1; break;
	case 'm': inmemory = 1; break;
	case 'M': showattributes = 1; break;
	case 'P': projection = strdup(optarg); break;
	case 'S': selection = strdup(optarg); break;
	case 'U': oclistpush(userparams,(ocelem)strdup(optarg)); break;
	case 'p':
	    if(strcasecmp(optarg,"das")==0) optdas=1;
	    else if(strcasecmp(optarg,"dds")==0) optdds=1;
	    else if(strcasecmp(optarg,"data")==0) optdatadds=1;
	    else if(strcasecmp(optarg,"datadds")==0) optdatadds=1;
	    else if(strcasecmp(optarg,"all")==0) {
		optdas = 1; optdds = 1; optdatadds = 1;
	    } else usage("unknown -p option");
	    break;
        default: usage("unknown option");
        }
    }

    if (debug > 0) {
        ocdebug = debug;
    }


    argc -= optind;
    argv += optind;

    if (argc > 0 && fileurl == NULL) {
        fileurl = strdup(argv[argc - 1]);
    }

    if (fileurl == NULL)
        fileurl = getenv("FILEURL");

    if (fileurl == NULL) {
        usage("no file url specified\n");
    }

    if (verbose)
        dumpflags();

    if (verbose) {
        fprintf(stdout, "initializing\n");
        fflush(stdout);
    }

    processdata(fileurl);

    return 0;
}

static void
dumpflags(void)
{
    fprintf(stderr,"url=%s\n",fileurl);
    if(verbose) fprintf(stderr,"verbose=%d\n",verbose);
    if(debug) fprintf(stderr,"debug=%d\n",debug);
    if(generate) fprintf(stderr,"generate=%d\n",generate);
    if(showattributes) fprintf(stderr,"showattributes=%d\n",showattributes);
    if(linear) fprintf(stderr,"linear=%d\n",linear);
    if(inmemory) fprintf(stderr,"inmemory=%d\n",inmemory);
    if(projection) fprintf(stderr,"projection=%s\n",projection);
    if(selection) fprintf(stderr,"selection=%s\n",selection);
    if(optdas || optdds || optdatadds) {
	fprintf(stderr,"-p options=");
	if(optdas) fprintf(stderr," das");
	if(optdds) fprintf(stderr," dds");
	if(optdatadds) fprintf(stderr," datadds");
	fprintf(stderr,"\n");
    }
    if(oclistlength(userparams) > 0) {
	unsigned int i;
	fprintf(stderr,"userparams=");
	for(i=0;i<oclistlength(userparams);i++) {
	    fprintf(stderr," %s",(char*)oclistget(userparams,i));
	}
	fprintf(stderr,"\n");
    }
}

static void
usage(char* msg)
{
    if(msg) fprintf(stderr,"error: %s\n",msg);
    fprintf(stderr,"usage: octest [-t][-v][-d][-M][-L][-D<level>][-m][-g][-p all|das|dds|data|datadds][-U <userparameter>] <fileurl>\n");
    fail(NULL);
}

static int
fail(char* msg)
{
    if(msg) fprintf(stderr,"fatalerror: %s\n",msg);
    fflush(stdout); fflush(stderr);
    exit(1);
}

/******************************************/
static void
check_err(int stat)
{
    if(stat == OC_NOERR) return;
    fprintf(stderr,"error status returned: (%d) %s\n",stat,ocerrstring(stat));
    fail(NULL);
}

static void
processdata(char* fileurl)
{
    int stat;
    OCconnection* state;
    OCnode* root;
    OCcontent* rootcontent;
    char* totalurl;
    unsigned int i;

    stat = oc_open(&state);
    check_err(stat);

    if(oclistlength(userparams) > 0) {
	OCbytes* buf = ocbytesnew();
	ocbytescat(buf,"[");
	for(i=0;i<oclistlength(userparams);i++) {
	    if(i>0) ocbytescat(buf,",");
	    ocbytescat(buf,(char*)oclistget(userparams,i));
	}
	ocbytescat(buf,"]");
	ocbytescat(buf,fileurl);
	totalurl = ocbytesdup(buf);
	ocbytesfree(buf);
    } else {
	totalurl = strdup(fileurl);
    }

    /* Always try to fetch metadata */
    stat = oc_fetchdas(state,totalurl);
    /* Ignore failure */

    stat = oc_fetchdds(state,totalurl);
    if(stat != OC_NOERR) fprintf(stderr,"Cannot read DDS: %s\n",fileurl);

    /* Read only if requested */
    if(optdatadds) {
        stat = oc_fetchdatadds(state,totalurl);
	if(stat != OC_NOERR) fprintf(stderr,"Cannot read DATADDS: %s\n",fileurl);
    }

    /* merge the das, if any*/
    if (oc_getdas(state) != NULL && oc_getdds(state) != NULL) {
	stat = oc_ddsdasmerge(state, MERGEDDS);
	check_err(stat);
    }

    if(generate) {
        if(optdas && oc_getdas(state) != NULL) {
	    fprintf(stdout,"%s",generatedas(state));
        }
        if(optdds && oc_getdds(state) != NULL) {
	    fprintf(stdout,"%s",generatedds(state));
        }
    } else {
        if(optdas && oc_getdastext(state)) {
	    fprintf(stdout,"%s",oc_getdastext(state));
        }
        if(optdds && oc_getddstext(state)) {
	    fprintf(stdout,"%s",oc_getddstext(state));
	}
    }

    if(!optdatadds) goto done;

    /* Insert user parameters */
    for(i=0;i<oclistlength(userparams);i++) {
	char* parm = strdup((char*)oclistget(userparams,i));
	char* value = strchr(parm,'=');
	if(value == NULL) {value = "";} else {*value = '\0'; value++;}
        dapurlreplace(&state->schema.dds.url,parm,value);
 	free(parm);
    }

    if(debug > 0) {
	fprintf(stderr,"Data packet dump:\n");
   	ocdumpfile(state->dap.file,state->dap.bod);
    }

    if(inmemory) {
	stat = oc_compile(state);
	check_err(stat);
	if(debug > 0) {
	    OCbytes* buf = ocbytesnew();
	    ocdumpmemdata(state->content.memdata,buf);
	    ocbytescat(buf,"");
	    fprintf(stderr,"OCmemdata Dump:\n%s",ocbytescontents(buf));
	    ocbytesfree(buf);
	}
    }

    /* Get the root content*/
    root = oc_getdatadds(state);
    rootcontent = oc_newcontent(state);
    stat = oc_rootcontent(state,rootcontent);
    check_err(stat);

    stacknext = 0;
    printleaves(state,root,rootcontent);

done:
    oc_close(state);

}

static void
printleaves(OCconnection* state, OCnode* node, OCcontent* content)
{
    int stat;
    unsigned int i,nsubnodes;
    OCbytes* buffer;

    buffer = ocbytesnew();

    if(linear) {
	OCcontent* fielddata  = oc_newcontent(state);
	unsigned int nsubnodes = oclistlength(node->subnodes);
        for(i=0;i<nsubnodes;i++) {
	    OCnode* subnode = (OCnode*)oclistget(node->subnodes,i);
	    ocbytescat(buffer,subnode->name);
	    ocbytescat(buffer," = ");
	    stat = oc_fieldcontent(state,content,fielddata,i);
	    check_err(stat);
	    printleavesr(state,subnode,fielddata,buffer);
            ocbytescat(buffer,"\n");
	}
	oc_freecontent(state,fielddata);
    } else {
        printleavesr(state,node,content,buffer);
    }
    fprintf(stdout,"%s",ocbytescontents(buffer));
}

static void
printleavesr(OCconnection* state, OCnode* node, OCcontent* content, OCbytes* buffer)
{
    int stat;
    unsigned int nsubnodes;
    unsigned int i,rank;
    unsigned int count;

    nsubnodes = oclistlength(node->subnodes);
    rank = node->array.rank;

    switch (oc_contentmode(state,content)) {
    case Fieldmode:
	{
	    OCcontent* fielddata  = oc_newcontent(state);
	    if(linear) ocbytescat(buffer,"{");
            for(i=0;i<nsubnodes;i++) {
	        OCnode* subnode = (OCnode*)oclistget(node->subnodes,i);
	        stat = oc_fieldcontent(state,content,fielddata,i);
	        check_err(stat);
		if(node->octype != OC_Dataset)
		    pushstack(subnode,Fieldmode,i,NULL);
	        if(i > 0) if(linear) ocbytescat(buffer,", ");
		printleavesr(state,subnode,fielddata,buffer);
		if(node->octype != OC_Dataset)
		    popstack();
	    }
	    oc_freecontent(state,fielddata);
	    if(linear) ocbytescat(buffer,"}");
	}
	break;

    case Dimmode:
        {
	    /* Walk to each element and dump*/
	    count = oc_dimcount(state,content);
	    if(linear) ocbytescat(buffer,"{");
	    if(node->octype == OC_Primitive) {
		size_t memsize = octypesize(node->etype)*count;
	        char* memory = malloc(memsize);
		size_t offset = 0;
		size_t avail = memsize;
	        for(i=0;i<count;i++) {
	            stat = oc_getcontent(state,content,(void*)(memory+offset),avail,i,1);
	            check_err(stat);
		    offset += octypesize(node->etype);
		    avail -= octypesize(node->etype);
		}
	        dumpinstance(node,count,memory,buffer);
		free(memory);
	    } else { /* structure*/
		OCcontent* dimdata = oc_newcontent(state);
		assert(node->octype == OC_Structure);
	        for(i=0;i<count;i++) {
	            stat = oc_dimcontent(state,content,dimdata,i);
	            check_err(stat);
		    pushstack(node,Dimmode,i,NULL);
		    if(i > 0) if(linear) ocbytescat(buffer,", ");
	            printleavesr(state,node,dimdata,buffer);
	            popstack();
		}
		oc_freecontent(state,dimdata);
	    }
	    if(linear) ocbytescat(buffer,"}");
	}
	break;

    case Recordmode:
	{
	    OCcontent* recorddata  = oc_newcontent(state);
            for(i=0;;i++) {/* could have used countrecords*/
	        stat = oc_recordcontent(state,content,recorddata,i);
		if(stat == OC_EINVALCOORDS) break; /* ith record does not exist*/
	        check_err(stat);
		pushstack(node,Recordmode,i,NULL);
		if(i > 0) if(linear) ocbytescat(buffer,", ");
		printleavesr(state,node,recorddata,buffer);
		popstack();
	    }
	    oc_freecontent(state,recorddata);
	}
	break;

    case Datamode:
	{ /* singleton data item*/
	    void* memory;
	    memory = calloc(octypesize(node->etype),1);
	    stat = oc_getcontent(state,content,memory,octypesize(node->etype),0,1);
	    check_err(stat);
	    dumpinstance(node,1,memory,buffer);
	    free(memory);
	}
	break;

    default: break;
    }
}

static void
dumpinstance(OCnode* node, size_t count, void* memory, OCbytes* buffer)
{
    unsigned int i;
    size_t delta = octypesize(node->etype);
    OCbytes* name;
    size_t max = count - 1;
    char tmp[1024];

    /* construct the name from the stack*/

    name = ocbytesnew();
    for(i=0;i<stacknext;i++) {
	struct DUMPPATH* entry = stack+i;
	ocbytescat(name,"/");
	ocbytescat(name,entry->node->name);
	switch (entry->mode) {
	case Dimmode: case Recordmode:
	    sprintf(tmp,"[%u]",entry->index);
	    ocbytescat(name,tmp);
	    break;
	case Fieldmode:
	    break;
	case Datamode:
	    break;
	default: ocbytescat(name,"?"); break;
 	}
    }
    if(!linear) {
	ocbytescat(buffer,"\t");
	snprintf(tmp,sizeof(tmp),"%s %s[%s%lu]",
		octypetostring(node->etype),
		ocbytescontents(name),
		(node->octype == OC_Primitive && max > 0)?"0..":"",
		(unsigned long)max);
	ocbytescat(buffer,tmp);
    }
    for(i=0;i<count;i++) {
	if(linear) {if(i > 0) ocbytescat(buffer,", ");}
	else ocbytescat(buffer," ");	
	octypeprint(node->etype,tmp,sizeof(tmp),(memory+(i*delta)));
	ocbytescat(buffer,tmp);	
    }
    if(!linear) ocbytescat(buffer,"\n");
    ocbytesfree(name);
    /* If the type was string, then free it up */
    if(node->etype == OC_String || node->etype == OC_URL) {
	char** strings = (char**)memory;
	for(i=0;i<count;i++) ocfree(strings[i]);
    }
}

static char*
generatedds(OCconnection* state)
{
    unsigned int i;
    char* result;
    OCbytes* buffer = ocbytesnew();
    char* esc;

    OCnode* root = oc_getdds(state);
    ocbytescat(buffer,"Dataset {\n");
    if(root != NULL)
    for(i=0;i<oclistlength(root->subnodes);i++) {
        OCnode* node = (OCnode*)oclistget(root->subnodes,i);
	generateddsr(state,node,buffer,1);
    }
    ocbytescat(buffer,"} ");
    if(root != NULL) ocbytescat(buffer,idescape(root->name));
    ocbytescat(buffer,";\n");
    result = ocbytesdup(buffer);
    ocbytesfree(buffer);
    return result;    
}

static char blanks[2048];
#define BLANKSPERDENT 2

static char*
indent(int n)
{
    int nblanks = BLANKSPERDENT * n;
    memset(blanks,' ',nblanks);
    blanks[nblanks] = '\0';
    return blanks;
}

static void
generateddsr(OCconnection* state, OCnode* node, OCbytes* buffer, int depth)
{
    unsigned int i;
    char tmp[256];
    unsigned int rank = node->array.rank;

    ocbytescat(buffer,indent(depth));

    switch (node->octype) {

    case OC_Primitive:
        snprintf(tmp,sizeof(tmp),"%s %s",octypetoddsstring(node->etype),idescape(node->name));
	ocbytescat(buffer,tmp);
	/* dump dim info*/
	for(i=0;i<rank;i++) {
	    OCnode* dim = (OCnode*)oclistget(node->array.dimensions,i);
	    if(dim->name == NULL)
	        snprintf(tmp,sizeof(tmp),"[%lu]",(unsigned long)dim->dim.declsize);
	    else
	        snprintf(tmp,sizeof(tmp),"[%s=%lu]",idescape(dim->name),(unsigned long)dim->dim.declsize);
	    ocbytescat(buffer,tmp);
	}
        ocbytescat(buffer,";\n");
	generateattributes(node,buffer,depth+1);
	break;

    case OC_Sequence:
        snprintf(tmp,sizeof(tmp),"Sequence {\n");
        ocbytescat(buffer,tmp);
	for(i=0;i<oclistlength(node->subnodes);i++) {
	    OCnode* subnode = (OCnode*)oclistget(node->subnodes,i);
	    generateddsr(state,subnode,buffer,depth+1);
	}
	snprintf(tmp,sizeof(tmp),"%s} %s;\n",indent(depth),idescape(node->name));
        ocbytescat(buffer,tmp);
	generateattributes(node,buffer,depth+1);
	break;

    case OC_Structure:
        snprintf(tmp,sizeof(tmp),"Structure {\n");
        ocbytescat(buffer,tmp);
	for(i=0;i<oclistlength(node->subnodes);i++) {
	    OCnode* subnode = (OCnode*)oclistget(node->subnodes,i);
	    generateddsr(state,subnode,buffer,depth+1);
	}
	snprintf(tmp,sizeof(tmp),"%s} %s",indent(depth),idescape(node->name));
	ocbytescat(buffer,tmp);
	for(i=0;i<rank;i++) {
	    OCnode* dim = (OCnode*)oclistget(node->array.dimensions,i);
	    if(dim->name == NULL)
	        snprintf(tmp,sizeof(tmp),"[%lu]",(unsigned long)dim->dim.declsize);
	    else
	        snprintf(tmp,sizeof(tmp),"[%s=%lu]",dim->name,(unsigned long)dim->dim.declsize);
	    ocbytescat(buffer,tmp);
	}
        ocbytescat(buffer,";\n");
	generateattributes(node,buffer,depth+1);
	break;

    case OC_Grid:
        snprintf(tmp,sizeof(tmp),"Grid {\n");
        ocbytescat(buffer,tmp);
	for(i=0;i<oclistlength(node->subnodes);i++) {
	    OCnode* subnode = (OCnode*)oclistget(node->subnodes,i);
	    if(i <= 1) {
	        ocbytescat(buffer,indent(depth+1));
	        if(i==0) 
		    ocbytescat(buffer,"Array:\n");
	        else
		    ocbytescat(buffer,"Maps:\n");
	    }
	    generateddsr(state,subnode,buffer,depth+2);
	}
	snprintf(tmp,sizeof(tmp),"%s} %s;\n",indent(depth),idescape(node->name));
        ocbytescat(buffer,tmp);
	generateattributes(node,buffer,depth+1);
	break;

    default:
	snprintf(tmp,sizeof(tmp),"ocget DDS: unexpected type: %d",node->octype);
	ocbytescat(buffer,tmp);
    }
}

static void
generateattributes(OCnode* node, OCbytes* buffer, int depth)
{
    unsigned int i,j;
    char tmp[128];
    unsigned int nattrs = oclistlength(node->attributes);

    if(showattributes && nattrs > 0) {
	for(i=0;i<nattrs;i++) {
	    OCattribute* att = (OCattribute*)oclistget(node->attributes,i);
	    char* value = att->values;
	    snprintf(tmp,sizeof(tmp),"%s%s %s:%s = ",indent(depth),
			octypetoddsstring(att->etype),idescape(node->name),att->name);
	    ocbytescat(buffer,tmp);
	    for(j=0;j<att->nvalues;j++) {
		if(j > 0) ocbytescat(buffer,", ");
		ocbytescat(buffer,
                           octypeprint(att->etype,tmp,sizeof(tmp),value));
		value += octypesize(att->etype);
	    }
	    ocbytescat(buffer,";\n");
	}
    }
}

static char*
generatedas(OCconnection* state)
{
    unsigned int i;
    char* result;
    OCbytes* buffer = ocbytesnew();

    OCnode* root = oc_getdas(state);
    ocbytescat(buffer,"Attributes {\n");
    for(i=0;i<oclistlength(root->subnodes);i++) {
        OCnode* node = (OCnode*)oclistget(root->subnodes,i);
	generatedasr(state,node,buffer,1);
    }
    ocbytescat(buffer,"}\n");
    result = ocbytesdup(buffer);
    ocbytesfree(buffer);
    return result;    
}

static void
generatedasr(OCconnection* state, OCnode* node, OCbytes* buffer, int depth)
{
    unsigned int i;
    char tmp[256];

    if(node->octype == OC_Attributeset) {
	snprintf(tmp,sizeof(tmp),"%s%s {\n",indent(depth),idescape(node->name));	
	ocbytescat(buffer,tmp);
	for(i=0;i<oclistlength(node->subnodes);i++) {
	    OCnode* attr = (OCnode*)oclistget(node->subnodes,i);
	    generatedasr(state,attr,buffer,depth+1);
	}
        ocbytescat(buffer,indent(depth));
        ocbytescat(buffer,"}\n");
    } else if(node->octype == OC_Attribute) {
	snprintf(tmp,sizeof(tmp),"%s%s %s",indent(depth),
		octypetoddsstring(node->etype),idescape(node->name));
	ocbytescat(buffer,tmp);
	for(i=0;i<oclistlength(node->att.values);i++) {
	    char* value = (char*)oclistget(node->att.values,i);
	    if(i > 0) ocbytescat(buffer,",");
	    if(node->etype == OC_String || node->etype == OC_URL)
	        snprintf(tmp,sizeof(tmp)," \"%s\"",stringescape(value));
	    else
	        snprintf(tmp,sizeof(tmp)," %s",value);
	    ocbytescat(buffer,tmp);
	}
	ocbytescat(buffer,";\n");
    } else {
	snprintf(tmp,sizeof(tmp),"ocget DAS: unexpected type: %d",node->octype);
	ocbytescat(buffer,tmp);
    }
}

static char hexdigits[] = "0123456789abcdef";

static char* escapedstring = NULL;

/* Add escape characters to a string */
static char*
stringescape(char* s)
{
    size_t len;
    char* p;
    int c;

    if(s == NULL) return NULL;
    len = strlen(s);
    if(escapedstring != NULL) free(escapedstring);
    escapedstring = (char*)malloc(4*len);
    p = escapedstring;
    while((c=*s++)) {
	if(c == '"' || c == '\\') {*p++ = '\\'; *p++ = c;}
	else if (c < ' ' || c >= 127) {
	    *p++ = '\\'; *p++ = 'x';
	    *p++ = hexdigits[(c & 0xf0)>>4];
	    *p++ = hexdigits[(c & 0xf)];
	} else
	    *p++ = c;
    }
    *p = '\0';
    return escapedstring;
}

static char idchars[] = "_.";
static char* escapedid = NULL;

/* Add escape characters to an identifier */
static char*
idescape(char* id)
{
    size_t len;
    char* p;
    int c;

    if(id == NULL) return NULL;
    len = strlen(id);
    if(escapedid != NULL) free(escapedid);
    escapedid = (char*)malloc(4*len);
    p = escapedid;
    while((c=*id++)) {
	if(c >= '0' && c <= '9') {*p++ = c;}
	else if(c >= 'a' && c <= 'z') {*p++ = c;}
	else if(c >= 'A' && c <= 'Z') {*p++ = c;}
	else if(strchr(idchars,c) != NULL) {*p++ = c;}
	else {
	    *p++ = '%';
	    *p++ = hexdigits[(c & 0xf0)>>4];
	    *p++ = hexdigits[(c & 0xf)];
	}
    }
    *p = '\0';
    return escapedid;
}
